/*  File FFOBJFN.C Feedforward neural net objective function.            */
/*                 This is the function to be minimized during training. */

#ifndef FFOBJFN_C
#define FFOBJFN_C

#include <conio.h>
#include <stdlib.h>
#include <mem.h>
#include <alloc.h>

#include "ut.h"
#include "trn.h"
#include "ffadv.h"

void network_wts_to_weights (ff_setup_rec setup,
  weights_rec network_weights,d_vector weights_vect);
void weights_to_network_wts (ff_setup_rec setup,
  d_vector weights_vect,weights_rec network_weights);
float ff_objective_fn (d_vector w);
void gradient_of_ff_obj_fn (d_vector w,d_vector grad);
float compute_max_obj_fn_value (ff_setup_rec setup,int no_of_items);

/* Implementation */

/*  Note: All weights arrays and vectors start with index 1  */

void network_wts_to_weights (ff_setup_rec setup,
     weights_rec network_weights,d_vector weights_vect) {
     int i,j,k = 0;
     for (j = 1;j<=setup.no_of_inputs + 1;j++)
	for (i = 1;i<=setup.no_of_h1_nodes;i++) {
	    k++;
	    weights_vect[k] = network_weights.wi[j][i];
	    } /* j,i */
     for (j = 1;j<=setup.no_of_h1_nodes;j++)
	 for (i = 1;i<=setup.no_of_outputs;i++) {
	    k++;
	    weights_vect[k] = network_weights.wo[j][i];
	    } /* j,i */
     return;
     }  /* end proc */

void weights_to_network_wts (ff_setup_rec setup,
     d_vector weights_vect,weights_rec network_weights) {
     int i,j,k = 0;
     for (j = 1;j<=(setup.no_of_inputs + 1);j++)
	 for (i = 1;i<=setup.no_of_h1_nodes;i++) {
	    k++;
	    network_weights.wi[j][i] = weights_vect[k];
	    } /* j,i */
     for (j = 1;j<=setup.no_of_h1_nodes;j++)
	 for (i = 1;i<=setup.no_of_outputs;i++) {
	     k++;
	     network_weights.wo[j][i] = weights_vect[k];
	     } /* j,i */
     return;
     } /* end proc */

float ff_objective_fn (d_vector w) {
  /* Note that g_tr_rec must be allocated and established */
  /*   before calling this procedure.   */
    int i;
    float sum = 0.0;
    network_rec network = {0,NULL,NULL,NULL};
    f_vector d_out = NULL;
    weights_rec network_wts = {0,NULL,NULL};
    if (!g_tr_rec.allocated) {
       cprintf ("\r\nff_objective_fn:  g_tr_rec not allocated.");
       getch();
       exit (1);
       }
    allocate_network (g_ff_setup,&network);
    network_wts = allocate_network_weights (g_ff_setup);
    d_out = allocate_f_vector (1,g_ff_setup.no_of_outputs);
    weights_to_network_wts (g_ff_setup,w,network_wts);
    for (i = 1;i<=g_tr_rec.no_of_training_items;i++) {
	memmove (network.x_inp + 1,g_tr_rec.training_array[i] + 1,
		g_ff_setup.no_of_inputs * sizeof(float));
	memmove (d_out + 1,g_tr_rec.norm_answers[i] + 1,
		g_ff_setup.no_of_outputs * sizeof(float));
	feedforward (network,network_wts,g_ff_setup);
	/*  Accumulate normalized error  */
	sum += ff_error (d_out,network.y_out,g_ff_setup);
	} /* i */
    if (heapcheck() < 0) {
       cprintf ("\r\n\nHeapcheck < 0 in objective_fn");
       getch();
       exit(1);
       }
    free_f_vector (d_out,1);
    free_weights (&network_wts);
    free_network (&network);
    return sum;
    }  /* end objective_fn */

void gradient_of_ff_obj_fn (d_vector w,d_vector grad) {
     int i,j,k,m,p,no_of_weights;
     float sum,factor;
     network_rec network = {0,NULL,NULL,NULL};
     f_vector d_out;
     weights_rec network_wts = {0,NULL,NULL};
     allocate_network (g_ff_setup,&network);
     network_wts = allocate_network_weights (g_ff_setup);
     d_out = allocate_f_vector (1,g_ff_setup.no_of_outputs);
     weights_to_network_wts (g_ff_setup,w,network_wts);
     no_of_weights = (g_ff_setup.no_of_inputs + 1) *
	    g_ff_setup.no_of_h1_nodes +
	    g_ff_setup.no_of_h1_nodes * g_ff_setup.no_of_outputs;
     for (i = 1;i<=no_of_weights;i++)
	 grad[i] = 0.0;
     for (p = 1;p<=g_tr_rec.no_of_training_items;p++) {
	 memmove (network.x_inp + 1,g_tr_rec.training_array[p] + 1,
		 g_ff_setup.no_of_inputs * sizeof(float));
	 memmove (d_out + 1,g_tr_rec.norm_answers[p] + 1,
		 g_ff_setup.no_of_outputs * sizeof(float));
	 feedforward (network,network_wts,g_ff_setup);
	 k = 0;
	 /* Input weights */
	 for (i = 1;i<=(g_ff_setup.no_of_inputs + 1);i++)
	     for (j = 1;j<=g_ff_setup.no_of_h1_nodes;j++) {
		k++;
		sum = 0.0;
		factor = network.h1[j] * (1.0 - network.h1[j])
			 * network.x_inp[i];
		for (m = 1;m<=g_ff_setup.no_of_outputs;m++)
		    sum += (d_out[m] - network.y_out[m])
			   * network.y_out[m] * (1.0 - network.y_out[m])
			   * network_wts.wo[j][m]
			   * factor;
		grad[k] += sum;
		} /* i,j */
	/* Output weights */
	for (j = 1;j<=g_ff_setup.no_of_h1_nodes;j++)
	    for (m = 1;m<=g_ff_setup.no_of_outputs;m++) {
		k++;
		grad[k] += (d_out[m] - network.y_out[m])
			   * network.y_out[m]*(1.0 - network.y_out[m])
			   * network.h1[j];
		} /* j,m */
	}  /* p */
    if (heapcheck() < 0) {
       cprintf ("\r\n\nHeapcheck < 0 in objective_fn");
       getch();
       exit(1);
       }
    free_f_vector (d_out,1);
    free_weights (&network_wts);
    free_network (&network);
    return;
    } /* proc */

float compute_max_obj_fn_value (ff_setup_rec setup,int no_of_items) {
    return (0.5 * no_of_items * setup.no_of_outputs *
      setup.norm_out_range * setup.norm_out_range);
    }

#endif